import java.io.*;
import java.net.*;

/**
 * The messenger is a mix of MessageSender and MessageReceiver from exercise 5.
 * The receiver part checks new packages and interprets them as commands for
 * this service. The sender part sends the answer for a service request back to
 * the user.
 * We could have used some inheritance/oo style, but it's simpler to
 * use copy & paste instead ;-).
 */
public class Messenger extends Thread {
	/** The quotes of the day. */
	protected static final String[] QOTD = {
		"If at first you don't succeed, failure may be your style.",
		"For every winner, there are dozens of losers. Odds are you're one of them.",
		"It could be that the purpose of your life is only to serve as a warning to others.",
		"Hard work often pays off after time, but laziness always pays off now.",
		"Never understimate the power of stupid people in large groups.",
		"The tallest blade of grass is the first to be cut by the lawnmower.",
		"In the battle between you and the world, bet on the world.",
		"There are no stupid questions, but there are a lot of inquisitive idiots.",
		"The best leaders inspire by example. When that's not an option, brute intimidation works pretty well, too.",
		"Sometimes the best solution to morale problems is just to fire all of the unhappy people.",
		"It's lonely at the top. But it's comforting to look down upon everyone at the bottom.",
		"If you never try anything, you'll miss out on many of life's great disappointments.",
		"I feel much better, now that I've given up hope.",
		"Why don't you write, and give me a chance not to reply?",
		"I want it clearly understood that I'm totally confused.",
		"Please ask your machine to get in touch with my machine.",
		"Keep smiling! (But not so much that people begin to wonder if you are mentally unbalanced.)"
	};
	protected int counter = 0;

	protected DatagramSocket socket = null;
	protected int port;


	/**
	 * Our constructor initializes the UDP socket.
	 */
	public Messenger(int port) {
		this.port = port;

		init();
	}

	/**
	 * Initializes the UDP socket. Stops the application, if something
	 * wents wrong.
	 */
	protected void init() {
		closeSocket(); // Might be necessary
		try {
			socket = new DatagramSocket(port);
		} catch (IOException ioe) {
			System.out.println("Oops, couldn't initialize UDP socket. Shutting down.");
			ioe.printStackTrace();
			System.exit(0); // Shut down.
		}
	}

	/**
	 * Our "destructor".
	 */
	protected void finalize() {
		closeSocket();
	}

	/**
	 * Just closes the socket.
	 */
	protected void closeSocket() {
		if (socket != null) {
			socket.close();
			socket = null;
		}
	}

	/**
	 * Almost equal to MessageReceiver.java in exercise 5.
	 * The message is interpreted as command for this service.
	 * If we receive 'help' as message text, we send a short
	 * description back.
	 * If we receive 'get qotd' as message text. we send the
	 * current Quote-of-the-day.
	 */
	public void run() {
		byte[] message = new byte[65536];

		while (true) {
			try {
				/* get the actual packet */
				DatagramPacket packet = new DatagramPacket(message,message.length);
				socket.receive(packet); // blocking, until packet received!
				int packetLength = packet.getLength();
				if (packetLength < 3) { // minimum length is 3, type (1) + messageLength (2)
					System.out.println("Unrecognized protocol format.");
					continue; // Ignore corrupt packets
				}

				/* Extract ip and port from the packet */
				InetAddress ip = packet.getAddress();
				int port = packet.getPort();

				/* parse the message */
				byte type = message[0];
				if (type != NetworkUtilities.MESSAGE) {
					System.out.println("Unrecognized protocol message.");
					continue; // Ignore corrupt packets
				}
				int messageLength = NetworkUtilities.get16(message,1);
				if (messageLength > packetLength - 3) { // Check for too long messages...
					messageLength = packetLength - 3;
				}
				String messageText = new String(message,3,messageLength); // Create a readable message as String

				/* Parsing the "command", eg. 'help' or 'get qotd' */
				if (messageText.equals("help")) { // User wants HELP
					sendHelp(ip,port);
					continue;
				}

				if (messageText.equals("get qotd")) { // User wants a QOTD
					sendQOTD(ip,port);
					continue;
				}

				sendHelp(ip,port); // Otherwise user is confused ;-), send help
			} catch (IOException ioe) {
				// I/O error in socket
				System.out.println("Oops, we caused an IOError.");
				ioe.printStackTrace();
				init(); // Fatal error, try to recover by initializing again...
			} catch (Exception e) {
				// other exceptions during parsing
				System.out.println("Other problems. Waiting for new message.");
				e.printStackTrace();
				continue; // Ignore any other exception
			}
		}
	}

	/**
	 * Sends a small help message to the user.
	 */
	protected static final String HELP = "I am a Quote-of-the-Day-Server.\nSend 'get qotd' to get my Quote-of-the-Day :-)";
	protected void sendHelp(InetAddress ip, int port) throws IOException {
		sendMessage(HELP,ip,port);
	}

	/**
	 * Sends the current Quote-of-the-Day to the user.
	 */
	protected void sendQOTD(InetAddress ip, int port) throws IOException {
		sendMessage(QOTD[counter],ip,port);
		counter = (counter + 1) % QOTD.length; // Setting the counter to the next QOTD...
	}

	/**
	 * Almost the same as in MessageSender.java, exercise 5.
	 */
    protected boolean sendMessage(String messageText, InetAddress ip, int port) throws IOException {
		try {
			/* Create packet (containing the messageText) and send it to the user */

			// Get some necessary information...
			int messageLength = messageText.length();
			byte[] message = new byte[3 + messageLength];

			// Create the message (as byte array)
			message[0] = NetworkUtilities.MESSAGE;
			NetworkUtilities.put16(messageLength,message,1);
			System.arraycopy(messageText.getBytes(),0,message,3,messageLength);

			// Create and send the UDP packet
			DatagramPacket packet = new DatagramPacket(message,message.length,ip,port);
			socket.send(packet);
		} catch (Exception e) {
			System.out.println("Oops, we caused an Error");
			e.printStackTrace();
			return false;
		}
		return true;
    }
}